/*
 * qsensview - Environment sensor data view
 *
 * Copyright (C) 2019  B. Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cmath>
#include "qsensview.h"
#include "qsvChartView.h"

#include <QDateTime>
#include <QPen>

qsvChartView::qsvChartView(float ymin, float ymax, const QString& unit, unsigned decis, QLabel *lbl, int yticks)
	: QChartView()
{
	decimals = decis;
	lbldata = lbl;
	miny = ymin;
	maxy = ymax;

	nsamples = pini->value("app/samples", 600).toInt();
	xquant = pini->value("app/xquant", 60).toInt() * 1000;
	int xticks = pini->value("app/xticks", 11).toInt();
	if(xticks < 3)
		xticks = 3;
	else if(xticks > 21)
		xticks = 21;
	if(nsamples < xticks)
		nsamples = xticks;
	else if(nsamples > 10000)
		nsamples = 10000;
	if(xquant < 60*1000)
		xquant = 60*1000;
	else if(xquant > 60*60*1000)
		xquant = 60*60*1000;

	data = new QLineSeries();
	mini = new QLineSeries();
	maxi = new QLineSeries();
	axisx = new QDateTimeAxis();
	axisy = new QValueAxis();
	chart = new QChart();

	mini->setPen(QPen(Qt::lightGray, 2));
	maxi->setPen(QPen(Qt::lightGray, 2));
	data->setPen(QPen(Qt::darkBlue, 3));

	chart->setMargins(QMargins(2, 2, 2, 2));
	chart->addSeries(mini);
	chart->addSeries(maxi);
	chart->addSeries(data);
	chart->legend()->hide();

	axisx = new QDateTimeAxis();
	axisx->setTickCount(xticks);
	axisx->setFormat("HH:mm");
	axisx->setGridLinePen(QPen(Qt::black));
	axisx->setMinorGridLinePen(QPen(Qt::black));

	axisy->setLabelFormat(QString("%.%1f").arg(decimals));
	axisy->setRange(ymin, ymax);
	axisy->setTickCount(abs(yticks));
	axisy->setMinorTickCount(yticks < 0 ? 0 : 1);
	axisy->setTitleText("[" + unit + "]");
	axisy->setGridLinePen(QPen(Qt::black));
	axisy->setMinorGridLinePen(QPen(QBrush(Qt::black), 1.0, Qt::DashLine));

	chart->addAxis(axisx, Qt::AlignBottom);
	chart->addAxis(axisy, Qt::AlignLeft);

	mini->attachAxis(axisx);
	mini->attachAxis(axisy);
	maxi->attachAxis(axisx);
	maxi->attachAxis(axisy);
	data->attachAxis(axisx);
	data->attachAxis(axisy);

	setChart(chart);
	setRenderHint(QPainter::Antialiasing);
}

void qsvChartView::setData(float val)
{
	lbldata->setText(QString::asprintf("%.*f", decimals, val));
}

QString qsvChartView::getDataText()
{
	return lbldata->text();
}


void qsvChartView::initData(unsigned interval)
{
	float val = (miny + maxy) / 2.0;
	qint64 now = QDateTime::currentMSecsSinceEpoch();
	for(int i = nsamples-1; i >= 0; i--) {
		mini->append(now - i*interval, val);
		maxi->append(now - i*interval, val);
		data->append(now - i*interval, val);
	}
	qint64 a = ::floor(data->at(0).x() / (float)xquant) * xquant;
	qint64 b = ::ceil(data->at(data->count()-1).x() / (float)xquant) * xquant;
	axisx->setRange(QDateTime::fromMSecsSinceEpoch(a), QDateTime::fromMSecsSinceEpoch(b));
}

void qsvChartView::addData(float val, float _min, float _max)
{
	if(val < miny)
		val = miny;
	else if(val > maxy)
		val = maxy;
	if(data->count() >= nsamples) {
		mini->remove(0);
		maxi->remove(0);
		data->remove(0);
	}
	qint64 now = QDateTime::currentMSecsSinceEpoch();
	if(data->count() < 1) {
		mini->append(now, val);
		maxi->append(now, val);
		data->append(now, val);
	} else {
		mini->append(now, _min);
		maxi->append(now, _max);
		data->append(now, val);
	}
	qint64 a = ::floor(data->at(0).x() / (float)xquant) * xquant;
	qint64 b = ::ceil(data->at(data->count()-1).x() / (float)xquant) * xquant;
	axisx->setRange(QDateTime::fromMSecsSinceEpoch(a), QDateTime::fromMSecsSinceEpoch(b));
}
