/*
 * qsensview - Environment sensor data view
 *
 * Copyright (C) 2019  B. Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "qsensview.h"
#include "qsvMainWindow.h"
#include "qsvChartView.h"

#include <QGridLayout>
#include <QVBoxLayout>
#include <QDateTime>
#include <QTimer>
#include <QTabWidget>

#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonValue>
#include <QHttpMultiPart>
#include <QHttpPart>

GraphView *qsvMainWindow::createGraph(int idx, QGridLayout *g, int col, int row)
{
	QString tag = QString("chart%1").arg(idx+1);
	if(!pini->contains(tag + "/label")) {
		printf("No tag %d '%s'\n", idx, tag.toUtf8().constData());
		return NULL;
	}

	QString label = pini->value(tag + "/label", "lbl").toString();
	QString units = pini->value(tag + "/units", "unit").toString();
	unsigned fsize = pini->value(tag + "/fsize", 12).toUInt();
	int nticks = pini->value(tag + "/yticks", 4).toInt();
	unsigned decimals = pini->value(tag + "/decimals", 0).toUInt();
	unsigned interval = pini->value("app/interval", 60).toUInt() * 1000;

	GraphView *gv = new GraphView;
	gv->tabindex = -1;	// Default to full display
	gv->jsontag    = pini->value(tag + "/jsontag", "???").toString();
	gv->dataindex  = pini->value(tag + "/dataindex", 0).toUInt();
	gv->datafactor = pini->value(tag + "/datafactor", 1.0).toDouble();
	gv->ymin = pini->value(tag + "/ymin", 0.0).toDouble();
	gv->ymax = pini->value(tag + "/ymax", 1.0).toDouble();

	gv->label = new QLabel(" " + label);
	gv->value = new QLabel("0");
	gv->units = new QLabel(" " + units + " ");

	gv->label->setFont(QFont("Sans", fsize));
	gv->value->setFont(QFont("Sans", fsize, QFont::Bold));
	gv->units->setFont(QFont("Sans", fsize));

	g->addWidget(gv->label, 2*row, 3*col+0);
	g->addWidget(gv->value, 2*row, 3*col+1, Qt::AlignRight);
	g->addWidget(gv->units, 2*row, 3*col+2);

	gv->view = new qsvChartView(gv->ymin, gv->ymax, units, decimals, gv->value, nticks);
	gv->view->initData(interval);
	g->addWidget(gv->view, 2*row+1, 3*col, 1, 3);

	return gv;
}

enum {
	BLS_ON,
	BLS_DIMMING,
	BLS_RISING,
	BLS_LOW,
};

static inline int clamp(int v, int mini, int maxi)
{
	if(v < mini)
		return mini;
	if(v > maxi)
		return maxi;
	return v;
}

qsvMainWindow::qsvMainWindow()
	: QWidget(0)
{
	firstData = true;
	datamissing = 0;

	sp = new QSerialPort(pini->value("app/port", "/dev/ttyUSB0").toString());
	sp->setBaudRate(pini->value("app/speed", 38400).toUInt());
	//sp->setDataBits(8);
	//sp->setStopBits(1);
	//sp->setParity(QSerialPort::NoParity);
	//sp->setFlowControl(QSerialPort::NoFlowControl);

	connect(sp, SIGNAL(readyRead()), this, SLOT(spRead()));
	connect(sp, SIGNAL(errorOccurred(QSerialPort::SerialPortError)), SLOT(spError(QSerialPort::SerialPortError)));
	if(!sp->open(QIODevice::ReadOnly)) {
		fprintf(stderr, "Cannot open serial port\n");
		exit(1);
	}

	setWindowTitle("Environment Sensors");
	int www = pini->value("app/wwidth", 1280).toInt();
	int hhh = pini->value("app/wheight", 720).toInt();
	resize(www, hhh);
	if(!pini->value("app/wcursor", true).toBool())
		setCursor(Qt::BlankCursor);

	blmin = clamp(pini->value("app/blmin", 16).toInt(), 0, 255);
	blmax = clamp(pini->value("app/blmax", 255).toInt(), 32, 255);
	blwait = clamp(pini->value("app/blwait", 30000).toInt(), 1000, 120000);
	blinterval = pini->value("app/blinterval", 0).toInt();
	if(blinterval)
		blinterval = clamp(blinterval, 25, 100);

	bltimer = new QTimer(this);
	bltimer->setSingleShot(true);
	connect(bltimer, SIGNAL(timeout()), this, SLOT(blTimeout()));
	blstate = BLS_ON;
	setBacklight(blight = blmax);
	activity();	// Starts the timer
	
	if(pini->value("app/wstyle", "plain").toString() == "tab") {
		tw = new QTabWidget();
		connect(tw, SIGNAL(currentChanged(int)), this, SLOT(tabEvent(int)));
		connect(tw, SIGNAL(tabBarClicked(int)), this, SLOT(tabEvent(int)));
		connect(tw, SIGNAL(tabBarDoubleClicked(int)), this, SLOT(tabEvent(int)));
		for(int i = 0; i < NGRAPHS; i++) {
			QGridLayout *g = new QGridLayout();
			g->setContentsMargins(QMargins(0, 0, 0, 0));
			g->setSpacing(0);
			graphs[i] = createGraph(i, g, 0, 0);
			graphs[i]->tabindex = i;
			QWidget *wg = new QWidget();
			wg->setLayout(g);
			tw->addTab(wg, graphs[i]->label->text());
		}
		QVBoxLayout *vb = new QVBoxLayout();
		vb->setContentsMargins(QMargins(0, 0, 0, 0));
		vb->setSpacing(0);
		vb->addWidget(tw);
		setLayout(vb);
	} else {
		tw = NULL;
		QGridLayout *g = new QGridLayout();
		g->setContentsMargins(QMargins(0, 0, 0, 0));
		g->setSpacing(0);

		for(int i = 0; i < NGRAPHS; i++) {
			graphs[i] = createGraph(i, g, i % 2, i / 2);
		}
		setLayout(g);
	}

	for(int i = 1; i < 25; i++) {
		QString dps = QString("data%1").arg(i);
		if(!(pini->contains(dps + "/jsontag") && pini->contains(dps + "/dataindex") && pini->contains(dps + "/datafactor")))
			continue;
		GraphView gv;
		gv.dataindex = pini->value(dps + "/dataindex", -1).toUInt();
		if(gv.dataindex >= 100)
			continue;
		gv.datafactor = pini->value(dps + "/datafactor", 1.0).toFloat();
		gv.jsontag = pini->value(dps + "/jsontag", "---").toString();
		if(gv.jsontag == "---")
			continue;
		gv.ymin = pini->value(dps + "/ymin", 0.0).toFloat();
		gv.ymax = pini->value(dps + "/ymax", 65535.0).toFloat();
		datapoints.append(gv);
	}

	setMouseTracking(true);

	connect(&qnam, SIGNAL(finished(QNetworkReply *)), this, SLOT(nwfinished(QNetworkReply *)));

	QTimer *t = new QTimer(this);
	connect(t, SIGNAL(timeout()), this, SLOT(timeout()));
	unsigned uv = pini->value("app/interval", 60).toUInt() * 1000;
	t->start(uv);
}

void qsvMainWindow::timeout()
{
	QJsonObject jdata;
	int ndata = 0;
	for(int i = 0; i < NGRAPHS; i++) {
		if(!graphs[i] || !graphs[i]->measurement.hasValues())
			continue;
		ndata++;
		float avg = graphs[i]->measurement.getAvg();
		float mini = graphs[i]->measurement.getMin();
		float maxi = graphs[i]->measurement.getMax();
		if(firstData) {
			// The min/max values are probably off because of data reset
			// Just go with the average
			mini = maxi = avg;
		}
		graphs[i]->view->addData(avg, mini, maxi);
		jdata.insert(graphs[i]->jsontag, QJsonObject({{"avg", avg}, {"min", mini}, {"max", maxi}}));
		graphs[i]->measurement.reset();
	}

	for(auto i = datapoints.begin(); i != datapoints.end(); i++) {
		if(!i->measurement.hasValues())
			continue;
		ndata++;
		float avg = i->measurement.getAvg();
		float mini = i->measurement.getMin();
		float maxi = i->measurement.getMax();
		if(firstData) {
			// The min/max values are probably off because of data reset
			// Just go with the average
			mini = maxi = avg;
		}
		jdata.insert(i->jsontag, QJsonObject({{"avg", avg}, {"min", mini}, {"max", maxi}}));
		i->measurement.reset();
	}

	// No actual data collected to send?
	if(!ndata) {
		datamissing++;
		// After 5 minutes of "no data", reset the arduino
		int resetmissing = pini->value("app/resetmissing", 5).toInt();
		if(datamissing >= resetmissing) {
			datamissing = 0;
			firstData = true;
			QString resetcmd = pini->value("app/resetcmd", "").toString();
			if(QFile::exists(resetcmd)) {
				QFileDevice::Permissions perm = QFile::permissions(resetcmd);
				if(perm & QFileDevice::ExeUser) {
					::system((resetcmd + " < /dev/null > /dev/null 2>&1").toUtf8().constData());
				}
			}
		}
		return;
	}

	datamissing = 0;
	if(!firstData) {
		// Send the data to the server
		QJsonObject jo;
		QString loc  = pini->value("app/location", "???").toString();
		QString key  = pini->value("app/apikey", "---").toString();
		QString ts   = QDateTime::currentDateTimeUtc().toString(Qt::ISODateWithMs);
		QByteArray hash = QCryptographicHash::hash((loc + ts + key).toUtf8(), QCryptographicHash::Sha256).toHex();
		jo.insert("Data",      jdata);
		jo.insert("Timestamp", ts);
		jo.insert("Location",  loc);
		jo.insert("Hash",      hash.constData());
		jo.insert("Interval",  (int)(pini->value("app/interval", 60).toUInt() * 1000));

		QJsonDocument jd(jo);
		QString uri = pini->value("app/restapi", "http://localhost/").toString();
		QUrl url(uri);
		QNetworkRequest req(url);
		QHttpMultiPart *mp = new QHttpMultiPart(QHttpMultiPart::FormDataType);
		QHttpPart jp;
		jp.setHeader(QNetworkRequest::ContentDispositionHeader, QVariant("form-data; name=\"json\""));
		jp.setHeader(QNetworkRequest::ContentTypeHeader, QVariant("application/json; charset=UTF-8"));
		jp.setBody(jd.toJson());
		mp->append(jp);
		QNetworkReply *reply = qnam.post(req, mp);
		mp->setParent(reply);
	} else {
		firstData = false;
	}
}

void qsvMainWindow::nwfinished(QNetworkReply *r)
{
	r->deleteLater();
}

void qsvMainWindow::spRead()
{
	spdata.append(sp->readAll());
	while(true) {
		int idx = spdata.indexOf('\n');
		if(idx < 0)
			break;
		QByteArray x = spdata.left(idx).trimmed();
		spdata.remove(0, idx+1);
		auto v = x.split(' ');
		// 'D 570.847 25.366 50.893 101683.484 25.330 13.879 0.000000 1469 1860'
		if(v.at(0) != "D")
			continue;

		for(int i = 0; i < NGRAPHS; i++) {
			if(!graphs[i])
				continue;
			float val;
			if((int)graphs[i]->dataindex < v.size()) {
				val = v.at(graphs[i]->dataindex).toFloat() * graphs[i]->datafactor;
			} else {
				val = (graphs[i]->ymin + graphs[i]->ymax) / 2.0;
			}
			graphs[i]->measurement.addValue(val);
			graphs[i]->view->setData(val);
			if(tw && graphs[i]->tabindex >= 0 && graphs[i]->tabindex < NGRAPHS) {
				tw->setTabText(graphs[i]->tabindex, graphs[i]->label->text() + " " + graphs[i]->view->getDataText());
			}
		}
		for(auto i = datapoints.begin(); i != datapoints.end(); i++) {
			float val;
			if((int)i->dataindex < v.size()) {
				val = v.at(i->dataindex).toFloat() * i->datafactor;
			} else {
				val = (i->ymin + i->ymax) / 2.0;
			}
			i->measurement.addValue(val);
		}
	}
}

void qsvMainWindow::spError(QSerialPort::SerialPortError e)
{
	if(e == QSerialPort::ReadError) {
		fprintf(stderr, "%s\n", QObject::tr("An I/O error occurred while reading "
			"the data from port %1, error: %2")
			.arg(sp->portName())
			.arg(sp->errorString()).toUtf8().constData());
	}
}

void qsvMainWindow::setBacklight(int val)
{
	if(!blinterval)
		return;

	if(val < blmin)
		val = blmin;
	else if(val > blmax)
		val = blmax;
	QFile qf(pini->value("app/bltarget", "/tmp/qsensview_key_bltarget_file_not_found").toString());
	if(qf.open(QIODevice::WriteOnly)) {
		qf.write(QString("%1\n").arg(val).toUtf8());
	} else {
		// File cannot be written, stop trying
		blinterval = 0;
	}
}

void qsvMainWindow::blTimeout()
{
	switch(blstate) {
	case BLS_DIMMING:
		if(blight > blmin) {
			blight--;
			bltimer->start(blinterval);
			setBacklight(blight);
		} else {
			blstate = BLS_LOW;
		}
		break;

	case BLS_RISING:
		if(blight < blmax) {
			blight += 10;
			if(blight > blmax)
				blight = blmax;
			bltimer->start(blinterval);
			setBacklight(blight);
		} else {
			blstate = BLS_ON;
			bltimer->start(blwait);
		}
		break;

	default:
	case BLS_ON:
		/* Wait fall timeout */
		blstate = BLS_DIMMING;
		bltimer->start(blinterval);
		break;

	case BLS_LOW:
		bltimer->stop();
		break;
	}
}

void qsvMainWindow::activity()
{
	if(!blinterval)
		return;

	switch(blstate) {
	case BLS_DIMMING:
	case BLS_LOW:
		blstate = BLS_RISING;
		bltimer->start(blinterval);
		break;

	case BLS_RISING:
		break;

	default:
		blstate = BLS_ON;
		blight = blmax;
		/* fallthrough */
	case BLS_ON:
		/* start/restart the timer on activity */
		bltimer->start(blwait);
		break;
	}
}

void qsvMainWindow::tabEvent(int)
{
	activity();
}

void qsvMainWindow::mouseDoubleClickEvent(QMouseEvent *)
{
	activity();
}

void qsvMainWindow::mouseMoveEvent(QMouseEvent *)
{
	activity();
}

void qsvMainWindow::mousePressEvent(QMouseEvent *)
{
	activity();
}

void qsvMainWindow::mouseReleaseEvent(QMouseEvent *)
{
	activity();
}

